#include <ReWindow/types/Window.h>
#include <ReWindow/Logger.h>
#include <vulkan/vulkan.h>
#include <vulkan/vulkan_xlib.h>
#include <set>

using namespace ReWindow;

VkInstance vulkan_context = nullptr;
VkApplicationInfo vulkan_application_info{};
VkSurfaceKHR vulkan_surface = nullptr;
VkDevice vulkan_logical_device = nullptr;
VkQueue vulkan_graphics_queue = nullptr;
VkQueue vulkan_present_queue = nullptr;

void VulkanWindow::Open() {
    xVars.display = XOpenDisplay(nullptr);
    xVars.defaultScreen = DefaultScreen(xVars.display);

    xVars.xSetWindowAttributes.border_pixel = BlackPixel(xVars.display, xVars.defaultScreen);
    xVars.xSetWindowAttributes.background_pixel = BlackPixel(xVars.display, xVars.defaultScreen);
    xVars.xSetWindowAttributes.override_redirect = True;
    xVars.xSetWindowAttributes.event_mask = ExposureMask;

    vulkan_application_info.sType = VK_STRUCTURE_TYPE_APPLICATION_INFO;
    vulkan_application_info.pApplicationName = title.c_str();
    vulkan_application_info.applicationVersion = VK_MAKE_VERSION(1, 0, 0);
    vulkan_application_info.pEngineName = "";
    vulkan_application_info.engineVersion = VK_MAKE_VERSION(1, 0, 0);

    vulkan_application_info.apiVersion = VK_API_VERSION_1_0;

    VkInstanceCreateInfo create_info{};
    create_info.sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO;
    create_info.pApplicationInfo = &vulkan_application_info;

    const char* extensions[] = { "VK_KHR_surface", "VK_KHR_xlib_surface" };
    create_info.enabledExtensionCount = 2;
    create_info.ppEnabledExtensionNames = extensions;
    if (vkCreateInstance(&create_info, nullptr, &vulkan_context) != VK_SUCCESS)
        throw std::runtime_error("Couldn't create Vulkan instance.");

    XVisualInfo visual_info_t{};
    visual_info_t.screen = xVars.defaultScreen;
    int visual_count;
    XVisualInfo* vi = XGetVisualInfo(xVars.display, VisualScreenMask, &visual_info_t, &visual_count);

    if (!vi)
        throw std::runtime_error("Failed to get X11 visual info.");

    Colormap colormap = XCreateColormap(xVars.display, RootWindow(xVars.display, vi->screen), vi->visual, AllocNone);
    xVars.xSetWindowAttributes.colormap = colormap;

    xVars.window = XCreateWindow(xVars.display, RootWindow(xVars.display, vi->screen),
                                 0, 0, width, height, 0, vi->depth, InputOutput,
                                 vi->visual, CWBackPixel | CWColormap | CWBorderPixel, &xVars.xSetWindowAttributes);

    if (!xVars.window)
        throw std::runtime_error("Failed to create X11 window for Vulkan.");

    XFree(vi);

    VkXlibSurfaceCreateInfoKHR surface_create_info{};
    surface_create_info.sType = VK_STRUCTURE_TYPE_XLIB_SURFACE_CREATE_INFO_KHR;
    surface_create_info.dpy = xVars.display;
    surface_create_info.window = xVars. window;

    if (vkCreateXlibSurfaceKHR(vulkan_context, &surface_create_info, nullptr, &vulkan_surface) != VK_SUCCESS)
        throw std::runtime_error("Failed to create Vulkan X11 surface.");

    unsigned int device_count = 0;
    vkEnumeratePhysicalDevices(vulkan_context, &device_count, nullptr);

    if (device_count == 0)
        throw std::runtime_error("Opening the window as Vulkan but there isn't a Vulkan compatible graphics card?");

    std::vector<VkPhysicalDevice> vulkan_devices(device_count);
    vkEnumeratePhysicalDevices(vulkan_context, &device_count, vulkan_devices.data());

    VkPhysicalDevice selected_device = VK_NULL_HANDLE;

    int graphics_family = -1;
    int present_family = -1;
    for (const auto& d : vulkan_devices) {
        unsigned int queue_family_count = 0;
        vkGetPhysicalDeviceQueueFamilyProperties(d, &queue_family_count, nullptr);

        std::vector<VkQueueFamilyProperties> queue_families(queue_family_count);
        vkGetPhysicalDeviceQueueFamilyProperties(d, &queue_family_count, queue_families.data());

        bool has_graphics = false;
        bool has_present = false;
        for (unsigned int i = 0; i < queue_family_count; i++) {
            if (queue_families[i].queueFlags & VK_QUEUE_GRAPHICS_BIT) { has_graphics = true; graphics_family = i; }

            VkBool32 present_support = false;
            vkGetPhysicalDeviceSurfaceSupportKHR(d, i, vulkan_surface, &present_support);
            if (present_support) { has_present = true; present_family = i; }

            if (has_graphics && has_present) { selected_device = d; break; }
        }
        if (selected_device != VK_NULL_HANDLE)
            break;
    }

    if (selected_device == VK_NULL_HANDLE)
        throw std::runtime_error("Opening the window using Vulkan but there isn't a Vulkan device which supports rendering & presenting?");

    VkPhysicalDeviceProperties device_properties;
    vkGetPhysicalDeviceProperties(selected_device, &device_properties);
    Logger::Debug("Vulkan continuing with selected device: " + std::string(device_properties.deviceName));

    float queue_prio = 1.0f;
    std::vector<VkDeviceQueueCreateInfo> queue_create_info;
    std::set<unsigned int> unique_queue_families = { (unsigned int) graphics_family, (unsigned int) present_family};

    for (unsigned int qf : unique_queue_families) {
        VkDeviceQueueCreateInfo vdqci{};
        vdqci.sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO;
        vdqci.queueFamilyIndex = qf;
        vdqci.queueCount = 1;
        vdqci.pQueuePriorities = &queue_prio;
        queue_create_info.push_back(vdqci);
    }

    // TODO allow the user to specify this.
    VkPhysicalDeviceFeatures enabled_features{};

    VkDeviceCreateInfo device_create_info{};
    device_create_info.sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO;
    device_create_info.queueCreateInfoCount = queue_create_info.size();
    device_create_info.pQueueCreateInfos = queue_create_info.data();
    device_create_info.pEnabledFeatures = &enabled_features;

    // TODO allow the user to specify this.
    std::vector<const char*> device_ext = {VK_KHR_SWAPCHAIN_EXTENSION_NAME};
    device_create_info.enabledExtensionCount = device_ext.size();
    device_create_info.ppEnabledExtensionNames = device_ext.data();

    if (vkCreateDevice(selected_device, &device_create_info, nullptr, &vulkan_logical_device) != VK_SUCCESS)
        throw std::runtime_error("Couldn't create the Vulkan logical device.");

    vkGetDeviceQueue(vulkan_logical_device, graphics_family, 0, &vulkan_graphics_queue);
    vkGetDeviceQueue(vulkan_logical_device, present_family, 0, &vulkan_present_queue);



    XSelectInput(xVars.display, xVars.window, ExposureMask | KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
                                              PointerMotionHintMask | FocusChangeMask | StructureNotifyMask | SubstructureRedirectMask | SubstructureNotifyMask | CWColormap );
    XMapWindow(xVars.display, xVars.window);
    XStoreName(xVars.display, xVars.window, title.c_str());

    xVars.windowTypeAtom = XInternAtom(xVars.display, "_NET_WM_WINDOW_TYPE", False);
    xVars.windowTypeUtilityAtom = XInternAtom(xVars.display, "_NET_WM_WINDOW_TYPE_UTILITY", False);
    XChangeProperty(xVars.display, xVars.window, xVars.windowTypeAtom, XA_ATOM, 32, PropModeReplace, (unsigned char*) &xVars.windowTypeUtilityAtom, 1);

    xVars.wmDeleteWindow = XInternAtom(xVars.display, "WM_DELETE_WINDOW", False);
    XSetWMProtocols(xVars.display, xVars.window, &xVars.wmDeleteWindow, 1);

    XGetWindowAttributes(xVars.display, xVars.window, &xVars.windowAttributes);
    render_area_position = { xVars.windowAttributes.x, xVars.windowAttributes.y };

    open = true;
    processOnOpen();
}
