#include <cstdlib>
#include <cstring>
#include <ReWindow/types/Window.h>
#include <ReWindow/types/Cursors.h>
#include <ReWindow/Logger.h>


ReWindow::IPair position = {0, 0};
bool should_poll_x_for_mouse_pos = true;

using namespace ReWindow;

void RWindow::Raise() {
    Logger::Debug(std::format("Raising window '{}'", this->title));
    // Get the position of the renderable area relative to the rest of the window.
    XGetWindowAttributes(xVars.display, xVars.window, &xVars.windowAttributes);
    render_area_position =  { xVars.windowAttributes.x, xVars.windowAttributes.y };
    XRaiseWindow(xVars.display, xVars.window);
}

void RWindow::Lower()
{
    Logger::Debug(std::format("Lowering window '{}'", this->title));
    XLowerWindow(xVars.display, xVars.window);
}

void RWindow::DestroyOSWindowHandle() {
    Logger::Debug(std::format("Destroying window '{}'", this->title));
    XDestroySubwindows(xVars.display, xVars.window);
    Logger::Debug(std::format("Destroyed window '{}'", this->title));
    XDestroyWindow(xVars.display, xVars.window);

    system("xset r on"); // Re-set X-server parameter to enable key-repeat.

    open = false;
}

//void RWindow::

void RWindow::SetCursorVisible(bool cursor_enable) {
    cursor_visible = cursor_enable;
    if (xVars.invisible_cursor == 0) {
        Pixmap blank_pixmap = XCreatePixmap(xVars.display, xVars.window, 1, 1, 1);
        XColor dummy; dummy.pixel = 0; dummy.red = 0; dummy.flags = 0;
        xVars.invisible_cursor = XCreatePixmapCursor(xVars.display, blank_pixmap, blank_pixmap, &dummy, &dummy, 0, 0);
        XFreePixmap(xVars.display, blank_pixmap);
    }

    if (!cursor_enable)
        XDefineCursor(xVars.display, xVars.window, xVars.invisible_cursor);

    if (cursor_enable)
        XUndefineCursor(xVars.display, xVars.window);
}

void RWindow::SetResizable(bool sizable) {
    XGetWindowAttributes(xVars.display, xVars.window, &xVars.windowAttributes);


    this->resizable = sizable;
    if (!sizable)
    {
        Logger::Debug("Once you've done this you cannot make it resizable again.");
        xVars.hints.flags = PMinSize | PMaxSize;
        xVars.hints.min_width = xVars.hints.max_width = xVars.windowAttributes.width;
        xVars.hints.min_height = xVars.hints.max_height = xVars.windowAttributes.height;
        XSetWMNormalHints(xVars.display, xVars.window, &xVars.hints);
    }
    //this->SetFlag(WindowFlag::RESIZABLE, resizable);

}

void RWindow::SetFlag(WindowFlag flag, bool state) {
    XGetWindowAttributes(xVars.display, xVars.window, &xVars.windowAttributes);
    flags[(int) flag] = state;
    //Once you've done this you cannot make it resizable again.
    if (flag == WindowFlag::RESIZABLE && !state) {
        Logger::Debug("Once you've done this you cannot make it resizable again.");
        xVars.hints.flags = PMinSize | PMaxSize;
        xVars.hints.min_width = xVars.hints.max_width = xVars.windowAttributes.width;
        xVars.hints.min_height = xVars.hints.max_height = xVars.windowAttributes.height;
        XSetWMNormalHints(xVars.display, xVars.window, &xVars.hints);
    }
    Logger::Debug(std::format("Set flag '{}' to state '{}' for window '{}'", WindowFlagToStr(flag), state, this->title));
}

void RWindow::PollEvents() {
    while(XPending(xVars.display)) {
        XNextEvent(xVars.display, &xVars.xev);

        if (xVars.xev.type == ClientMessage)
            Logger::Info(std::format("Event '{}'", "ClientMessage"));

        if (xVars.xev.xclient.message_type == XInternAtom(xVars.display, "WM_PROTOCOLS", False) && static_cast<Atom>(xVars.xev.xclient.data.l[0]) == xVars.wmDeleteWindow) {
            Close();
        }

        if (xVars.xev.type == FocusIn) {
            Logger::Debug(std::format("Event'{}'", "FocusIn"));
            XAutoRepeatOff(xVars.display);
            SetFlag(WindowFlag::IN_FOCUS, true);

            if (!cursor_visible)
                XDefineCursor(xVars.display, xVars.window, xVars.invisible_cursor);

            // Get the position of the renderable area relative to the rest of the window.
            XGetWindowAttributes(xVars.display, xVars.window, &xVars.windowAttributes);
            render_area_position =  { xVars.windowAttributes.x, xVars.windowAttributes.y };
            processFocusIn();
        }

        if (xVars.xev.type == FocusOut) {
            Logger::Debug(std::format("Event '{}'", "FocusOut"));
            XAutoRepeatOn(xVars.display);

            SetFlag(WindowFlag::IN_FOCUS, false);
            if (!cursor_visible)
                XUndefineCursor(xVars.display, xVars.window);

            processFocusOut();
        }

        if (xVars.xev.type == KeyRelease) {
            Logger::Debug(std::format("Event '{}'", "KeyRelease"));
            auto scancode = (X11Scancode) xVars.xev.xkey.keycode;
            auto key = GetKeyFromX11Scancode(scancode);
            processKeyRelease(key);
        }

        if (xVars.xev.type == KeyPress) {
            Logger::Debug(std::format("Event '{}'", "KeyPress"));
            auto scancode = (X11Scancode) xVars.xev.xkey.keycode;
            auto key = GetKeyFromX11Scancode(scancode);
            processKeyPress(key);
        }

        if (xVars.xev.type == ButtonRelease) {

            // Mouse Wheel fires both the ButtonPress and ButtonRelease instantaneously.
            // Therefore, we handle it as a specific MouseWheel event rather than a MouseButton event,
            // and only call on ButtonPress, otherwise it will appear to duplicate the mouse wheel scroll.
            if (xVars.xev.xbutton.button != 4 && xVars.xev.xbutton.button != 5) {
                MouseButton button = GetMouseButtonFromXButton(xVars.xev.xbutton.button);
                Logger::Debug(std::format("Event '{}'", "ButtonRelease"));
                processMouseRelease(button);
            }
        }

        if (xVars.xev.type == ButtonPress) {

            // Mouse Wheel fires both the ButtonPress and ButtonRelease instantaneously.
            // Therefore, we handle it as a specific MouseWheel event rather than a MouseButton event,
            // and only call on ButtonPress, otherwise it will appear to duplicate the mouse wheel scroll.
            if (xVars.xev.xbutton.button == 4) {
                processMouseWheel(-1);
            } else if (xVars.xev.xbutton.button == 5) {
                processMouseWheel(1);
            } else
            {
                MouseButton button = GetMouseButtonFromXButton(xVars.xev.xbutton.button);

                Logger::Debug(std::format("Event: MouseButtonPress {}", button.Mnemonic));

                processMousePress(button);
            }
        }

        if (xVars.xev.type == Expose)
        {
            Logger::Debug(std::format("Event '{}'", "Expose"));
        }

        // NOTE: This event is functionally useless, as it only informs of the very beginning and end of a mouse movement.
        if (xVars.xev.type == MotionNotify)
        {
            Logger::Debug(std::format("Event '{}'", "MotionNotify"));
        }

        if (xVars.xev.type == ConfigureNotify) {
            if (this->width != xVars.xev.xconfigurerequest.width || this->height != xVars.xev.xconfigurerequest.height) {
                Logger::Debug(std::format("Event '{}'", "ResizeRequest"));

                this->width = xVars.xev.xconfigurerequest.width;
                this->height = xVars.xev.xconfigurerequest.height;

                auto eventData = WindowResizeRequestEvent();
                eventData.Size = { xVars.xev.xconfigurerequest.width, xVars.xev.xconfigurerequest.height };
                lastKnownWindowSize = eventData.Size;

                OnResizeRequest(eventData);
                OnResizeRequestEvent(eventData);
            }

            //Window Moved.
            if (position.x != xVars.xev.xconfigurerequest.x || position.y != xVars.xev.xconfigurerequest.y)
                position = { xVars.xev.xconfigurerequest.x, xVars.xev.xconfigurerequest.y };
        }

    }
    previousKeyboard = currentKeyboard;
    previousMouse.Buttons = currentMouse.Buttons;
}


// Might make the window go off the screen on some window managers.
void RWindow::SetSize(int newWidth, int newHeight) {
    if (!resizable) return;

    this->width = newWidth;
    this->height = newHeight;
    XResizeWindow(xVars.display, xVars.window, newWidth, newHeight);
    XFlush(xVars.display);
    Logger::Info(std::format("Set size for '{}' to {} x {}", this->title, newWidth, newHeight));
}

IPair RWindow::GetAccurateMouseCoordinates() const {

    Window root_return, child_return;
    int root_x_ret, root_y_ret;
    int win_x_ret, win_y_ret;
    uint32_t mask_return;

    // This seems to be relative to the top left corner of the renderable area.
    bool mouseAvailable = XQueryPointer(xVars.display, xVars.window, &root_return, &child_return, &root_x_ret, &root_y_ret, &win_x_ret, &win_y_ret, &mask_return);

    if (mouseAvailable) {
        // TODO: normalize coordinates from xVars.displaySpace to windowSpace
        // TODO: fire mouse movement event
        IPair m_coords = { win_x_ret, win_y_ret };
        return m_coords;
    }
    return {};
}


IPair RWindow::GetSize() const {
    return { this->width, this->height};
}

//IPair RWindow::getLastKnownResize() const
//{
//    return lastKnownWindowSize;
//}

// TODO: implement integer IPair/3 types
IPair RWindow::GetPos() const {
    return position;
}

void RWindow::SetPos(int x, int y) {
    XMoveWindow(xVars.display, xVars.window, x, y);
    position = {x, y};
}

void RWindow::SetPos(const IPair& pos) {
    SetPos(pos.x, pos.y);
}

void RWindow::Fullscreen() {
    Logger::Info(std::format("Fullscreening '{}'", this->title));
    fullscreen_mode = true;

    XEvent xev;
    Atom wm_state   = XInternAtom(xVars.display, "_NET_WM_STATE", true);
    Atom wm_fullscreen = XInternAtom(xVars.display, "_NET_WM_STATE_FULLSCREEN", true);

    XChangeProperty(xVars.display, xVars.window, wm_state, XA_ATOM, 32, PropModeReplace, (unsigned char *)&wm_fullscreen, 1);
    memset(&xev, 0, sizeof(xev));
    xev.type = ClientMessage;
    xev.xclient.window = xVars.window;
    xev.xclient.message_type = wm_state;
    xev.xclient.format = 32;
    xev.xclient.data.l[0] = 1; // _NET_WM_STATE_ADD
    xev.xclient.data.l[1] = fullscreen_mode;
    xev.xclient.data.l[2] = 0;
    XSendEvent(xVars.display, DefaultRootWindow(xVars.display), False, SubstructureNotifyMask | SubstructureRedirectMask, &xev);
    Logger::Debug(std::format("Fullscreened '{}'", this->title));
}

void RWindow::RestoreFromFullscreen() {
    Logger::Debug(std::format("Restoring '{}' from Fullscreen", this->title));
    fullscreen_mode = false;
    XEvent xev;
    Atom wm_state = XInternAtom(xVars.display, "_NET_WM_STATE", False);
    Atom fullscreen = XInternAtom(xVars.display, "_NET_WM_STATE_FULLSCREEN", False);
    memset(&xev, 0, sizeof(xev));
    xev.type = ClientMessage;
    xev.xclient.window = xVars.window;
    xev.xclient.message_type = wm_state;
    xev.xclient.format = 32;
    xev.xclient.data.l[0] = 0; // _NET_WM_STATE_REMOVE
    xev.xclient.data.l[1] = fullscreen_mode;
    xev.xclient.data.l[2] = 0;
    XSendEvent(xVars.display, DefaultRootWindow(xVars.display), False, SubstructureNotifyMask | SubstructureRedirectMask, &xev);
    Logger::Debug(std::format("Restored '{}' from Fullscreen", this->title));
}

void RWindow::SetCursorStyle(CursorStyle style) const {
    auto x11_cursor_resolved_enum = (unsigned int) style.X11Cursor;
    Cursor c = XCreateFontCursor(xVars.display, x11_cursor_resolved_enum);
    XDefineCursor(xVars.display, xVars.window, c);
}

void RWindow::SetTitle(const std::string& title) {
    this->title = title;
    XStoreName(xVars.display, xVars.window, title.c_str());
}



IPair RWindow::GetPositionOfRenderableArea() const {
    return render_area_position;
}

